from enum import *
from formule import *

# On parse nos formules a la main.
# On fait une analyse lexicale, puis syntaxique.

# Voici la grammaire de nos formules (les crochets [] sont a ecrire) :
#    formule -> T | B | AP | connecteur1 [formule U formule] |connecteur1 [formule W formule] | connecteur1 [connecteur2 formule] | quantificateur.varList[formule] | !formule | formule connecteurBool formule | E[macro1 macro2 int formule]
#    connecteur1 -> A | E
#    connecteur2 -> X | F | G
#    connecteurBool -> & | '|' (caractere pipe)
#    quantificateur -> Ex | All | EX1 
#    AP -> Un identificateur de variable. Il faudrait eviter qu'il contienne une majuscule pour eviter les ambiguites.
#    varList -> AP | AP,varList
#    macro1 -> X | F
#    macro2 -> > | =
#    int -> Un entier positif.

# Les espaces et parentheses standard () sont autorises (et parfois necessaires).
# T (top) represente la tautologie et B (bottom) represente l'absurde.
# EX>k f (resp. EX=k f) signifie qu'il existe au moins (resp. exactement) k successeurs qui satisfont f.
# De meme avec EF pour les accessibles.
class Token(Enum) :
    NOT=1
    AND=2
    OR=3
    E=4
    A=5
    F=6
    G=7
    U=8
    W=9
    X=10
    LBRACK=11
    RBRACK=12
    LPAR=13
    RPAR=14
    B=15
    T=16
    SUP=17
    EQ=18
    COMMA=19

# Prends une chaine de caracteres, y lit une formule, et renvoie la liste
# des noeuds de l'AST de la formule dans l'ordre suffixe.
def lex(s) :
    ap="" # On lit caractere par caractere, ap est donc une potentielle proposition atomique qu'on met a jour a chaque caractere lu.
    stack=[] # pour stocker les connecteurs (on retourne une representation suffixe)
    ans=[]
    # Certains connecteurs unaires sont prioritaires.
    # Par exemple, dans la formule, !a & b,
    # la negation est sur a et non sur a&b.
    # popNot permet de decider a qui s'appliquent les connecteur unaires
    # quand un connecteur infixe est push dans la pile.
    def popNot() :
        while (stack != []) and ((stack[-1] == Token.NOT) or isinstance(stack[-1], FORALL) or isinstance(stack[-1], EXISTS) or isinstance(stack[-1], EXISTS1) or isinstance(stack[-1], FORALL1)  ) :
            ans.append(stack.pop())
    # On a atteint la fin d'un mot (un caractere d'espacement, une parenthese, un connecteur...).
    # ap a ete construit, et on doit maintenant le push.
    def pushAP(ap) :
        if ap == "E" :
            stack.append(Token.E)
        elif ap == "A" :
            stack.append(Token.A)
        elif ap != "" :
            if (stack != []) and ((isinstance(stack[-1], FORALL)) or (isinstance(stack[-1], EXISTS)) or (isinstance(stack[-1], EXISTS1)) or (isinstance(stack[-1], FORALL1))) :
                stack[-1].var.append(ap)
            else :
                ans.append(AP(ap))
        return ""
    for car in s :
        if car == "!" :
            ap=pushAP(ap)
            stack.append(Token.NOT)

        elif car == "." :
            if ap == "All" :
                stack.append(FORALL([], None))
            elif ap == "All1" :
                stack.append(FORALL1([], None))
            elif ap == "Ex" :
                stack.append(EXISTS([], None))
            elif ap == "Ex1" :
                stack.append(EXISTS1([], None))
            ap = ""

        elif car == "&" :
            ap=pushAP(ap)
            popNot()
            stack.append(Token.AND)

        elif car == "|" :
            ap=pushAP(ap)
            popNot()
            stack.append(Token.OR)

        elif car == "X" :
            ap=pushAP(ap)
            stack.append(Token.X)

        elif car == "F" :
            ap=pushAP(ap)
            stack.append(Token.F)

        elif car == "G" :
            ap=pushAP(ap)
            stack.append(Token.G)

        elif car == "U" :
            ap=pushAP(ap)
            popNot()
            stack.append(Token.U)
            
        elif car == "W" :
            ap=pushAP(ap)
            popNot()
            stack.append(Token.U)

        elif car == "(" :
            ap=pushAP(ap)
            stack.append(Token.LPAR)

        elif car == ")" :
            ap=pushAP(ap)
            while stack[-1] != Token.LPAR :
                ans.append(stack.pop())
            stack.pop() # on jette la parenthese gauche

        elif car == "[" :
            ap=pushAP(ap)
            stack.append(Token.LBRACK)

        elif car == "]" :
            ap=pushAP(ap)
            while stack[-1] != Token.LBRACK :
                ans.append(stack.pop())
            stack.pop() # on jette le crochet droit
            ans.append(stack.pop()) # le sommet de stack est soit un quantificateur, soit un quantificateur d'etat (E ou A)

        elif car.isspace() :
            ap=pushAP(ap)

        elif car == "B" :
            ans.append(BOTTOM())

        elif car == "T" :
            ans.append(TOP())

        elif car == ">" :
            ap=pushAP(ap)
            stack.append(Token.SUP)

        elif car == "=" :
            ap=pushAP(ap)
            stack.append(Token.EQ)

        elif car == "," :
            ap=pushAP(ap)
            
        else :
            ap = ap + car
    while stack != [] :
        ans.append(stack.pop())
    return ans

# On a la suite de tokens en representation suffixe, et on veut calculer l'AST.
def computeFormula(tokenStream) :
    stack=[]
    # en represantation suffixe, les arguments des fonctions
    # sont ecrit avant celles-ci, on les stocke donc dans une pile
    tmp=None # par rapport a la grammaire definie en haut, tmp est le connecteur2 en attendant de lire le connecteur1
    for token in tokenStream :
        if token == Token.NOT :
            stack.append(NOT(stack.pop()))

        elif token == Token.AND :
            tmp=stack.pop()
            stack.append(AND(stack.pop(), tmp))

        elif token == Token.OR :
            tmp=stack.pop()
            stack.append(OR(stack.pop(), tmp))

        elif token == Token.E :
            if tmp == Token.U :
                tmp=stack.pop()
                stack.append(EU(stack.pop(), tmp))

            elif tmp == Token.W :
                tmp=stack.pop()
                stack.append(EW(stack.pop(), tmp))

            elif tmp == Token.X :
                stack.append(EX(stack.pop()))

            elif tmp == Token.F :
                stack.append(EF(stack.pop()))

            elif tmp == Token.G :
                stack.append(EG(stack.pop()))

            elif tmp == (Token.X, Token.SUP) :
                prop = stack.pop()
                stack.append(getXAtLeast(int(stack.pop().ap)+1, prop))

            elif tmp == (Token.F, Token.SUP) :
                prop = stack.pop()
                stack.append(getFAtLeast(int(stack.pop().ap)+1, prop))

            elif tmp == (Token.X, Token.EQ) :
                prop = stack.pop()
                stack.append(getXExactly(int(stack.pop().ap), prop))

            elif tmp == (Token.F, Token.EQ) :
                prop = stack.pop()
                stack.append(getFExactly(int(stack.pop().ap), prop))

        elif token == Token.A :
            if tmp == Token.U :
                tmp=stack.pop()
                stack.append(AU(stack.pop(), tmp))
                
            elif tmp == Token.W :
                tmp=stack.pop()
                stack.append(AW(stack.pop(), tmp))

            elif tmp == Token.X :
                stack.append(AX(stack.pop()))

            elif tmp == Token.F :
                stack.append(AF(stack.pop()))

            elif tmp == Token.G :
                stack.append(AG(stack.pop()))

        elif token == Token.U :
            tmp = token

        elif token == Token.X :
            if tmp == Token.SUP or tmp == Token.EQ :
                tmp = (token, tmp)
            else :
                tmp = token

        elif token == Token.F :
            if tmp == Token.SUP or tmp == Token.EQ :
                tmp = (token, tmp)
            else :
                tmp = token
            
        elif token == Token.G :
            tmp = token

        elif isinstance(token, FORALL) or isinstance(token, EXISTS)  or isinstance(token, EXISTS1)  or isinstance(token, FORALL1):
            token.fils=[stack.pop()]
            stack.append(token)

        elif token == Token.SUP or token == Token.EQ :
            tmp=token

        else :#AP, BOTTOM, TOP
            stack.append(token)

    return stack.pop()

# De string vers AST.
def formula_of_string(s) :
    return computeFormula(lex(s))

# De fichier vers AST.
def read_formula(pathname) :
    input=open(pathname, "r")
    s=input.read()
    input.close()
    return formula_of_string(s)
